import argparse
import sys

from .. import freeze_to_string

from http.server import BaseHTTPRequestHandler, HTTPServer
import socketserver

class FrozenSoupRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        url = self.path[1:] # trim leading /

        # TODO Some error handling would be nice
        content = freeze_to_string(url)

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(bytes(content, 'utf-8'))

def main() -> int:
    parser = argparse.ArgumentParser(
        prog = 'server',
        description  = 'Serve up single-file HTML page for a URL',
    )
    parser.add_argument(
        '-H', '--host',
        type=str,
        default='localhost',
        help='default port'
    )
    parser.add_argument(
        '-p', '--port',
        type=int,
        default=8080,
        help='default port'
    )

    args = parser.parse_args()

    with socketserver.TCPServer((args.host, args.port), FrozenSoupRequestHandler) as httpd:
        print(f"Server started http://{args.host}:{args.port}")
        httpd.serve_forever()

    return 0

if __name__ == '__main__':
    sys.exit(main())



from typing import Optional, Union

import requests
import tinycss2

from ..resource import get_ref_as_dataurl

def expand_urls_in_css(
    css: str,
    base_url: str,
    session: Optional[requests.Session] = None,
    timeout: Union[float, tuple[float, float], None] = 900.0,
) -> str:
    rules = tinycss2.parse_stylesheet(css)

    output = ""

    for rule in rules:
        transformed = expand_urls_in_rule(rule, base_url, session, timeout)
        output += transformed.serialize()

    return output

def expand_urls_in_rule(
    rule,
    base_url: str,
    session: Optional[requests.Session] = None,
    timeout: Union[float, tuple[float, float], None] = 900.0,
):
    if type(rule) == tinycss2.ast.QualifiedRule or type(rule) == tinycss2.ast.AtRule:
        for child_rule in rule.content:
            child_rule = expand_urls_in_rule(child_rule, base_url, session, timeout)
        return rule
    elif type(rule) == tinycss2.ast.URLToken:
        data_url = get_ref_as_dataurl(base_url, rule.value, session, timeout)
        rule.representation = f'url({data_url})'
        return rule
    else:
        return rule



from typing import Optional, Union, List

import requests

from bs4 import BeautifulSoup

from urllib.parse import urljoin
import base64

from .resource import get_ref_as_dataurl
from .css import expand_urls_in_css

def freeze_to_string(
    url: str,
    session: Optional[requests.Session] = None,
    timeout: Union[float, tuple[float, float], None] = 900.0,
    formatter: str = 'html5',
    knockouts: Optional[List[str]] = None,
) -> str:
    if session is None:
        session = requests.Session()

    r = session.get(url, timeout= timeout)

    soup = BeautifulSoup(r.text, 'html.parser')

    # Process the knockouts first so we don't do any extra work on those
    if knockouts is not None:
        for selector in knockouts:
            for tag in soup.css.select(selector):
                tag.decompose()

    base_url = url

    # Find the first <base href="">, which could follow a <base target="">
    for base_tag in soup.find_all('base'):
        base_href = base_tag.get('href')
        if base_href is not None:
            base_url = urljoin(url, base_href)
            break

    # Inline images
    for img in soup.find_all('img'):
        img['src'] = get_ref_as_dataurl(base_url, img['src'], session, timeout)

    # Handle <link> elements
    for link in soup.find_all('link'):
        # Inline rel="icon"
        if 'icon' in link.get_attribute_list('rel'):
            link['href'] = get_ref_as_dataurl(base_url, link['href'], session, timeout)

        elif 'apple-touch-icon' in link.get_attribute_list('rel'):
            link['href'] = get_ref_as_dataurl(base_url, link['href'], session, timeout)

        elif 'apple-touch-startup-image' in link.get_attribute_list('rel'):
            link['href'] = get_ref_as_dataurl(base_url, link['href'], session, timeout)

        # Turn rel="stylesheet" into <style>
        elif 'stylesheet' in link.get_attribute_list('rel'):
            stylesheet_url = urljoin(base_url, link['href'])
            response = session.get(stylesheet_url, timeout=timeout)
            if response.status_code == 200:
                style = soup.new_tag('style')
                style.string = expand_urls_in_css(response.text, stylesheet_url, session, timeout)
                # Carry over media=""
                if link.get('media'):
                    style['media'] = link['media']
                # TODO anything else?
                link.replace_with(style)
            else:
                raise Exception(f"Unable to replace style {link['href']}")

    # Inline <script src="">
    for script in soup.find_all('script'):
        if script.get('src'):
            response = session.get(urljoin(base_url, script['src']), timeout=timeout)
            if response.status_code == 200:
                script.string = response.text
                # TODO what other attributes do we care about?
                # TODO parse/rewrite JavaScript to handle `import`?
                del script['src']
            else:
                raise Exception(f"Unable to replace script contents {script['src']}")

    # Should allow the caller to specify the formatter to use, html5 for now
    return soup.decode(formatter=formatter)



from typing import Optional, Union

import base64
import requests
from urllib.parse import urljoin

def get_ref_as_dataurl(
    base_url: str,
    ref_url: str,
    session: requests.Session,
    timeout: Union[float, tuple[float, float], None] = 900.0,
) -> str:
    url = urljoin(base_url, ref_url)

    # If it's already a data: URL, we're okay
    if url.startswith('data:'):
        return url

    response = session.get(url, timeout= timeout)
    response.raise_for_status()

    # Encode the content to base64 - don't use urlsafe_b64encode because
    # the browser won't be passing this on to a server, and they don't
    # expect the 'URL-safe' substitutions
    encoded_content = base64.b64encode(response.content).decode("utf-8")

    # Grab the content-type from the response headers
    # TODO: what if we have no content_type?
    content_type = response.headers.get('Content-Type')

    # Return the data: URL with appropriate MIME type
    return f"data:{content_type};base64,{encoded_content}"



import sys
import os

import argparse

from . import freeze_to_string

# Based on what pip does, but try to be brief and not hardcode our name
def get_prog() -> str:
    package = __loader__.name.split(".")[0]
    try:
        prog = os.path.basename(sys.argv[0])
        if prog in ("__main__.py", "-c"):
            # go back to orig_argv[0] to get what the user used
            return f"{sys.orig_argv[0]} -m {package}"
        else:
            return prog
    except (AttributeError, TypeError, IndexError):
        pass
    return package

def main() -> int:
    parser = argparse.ArgumentParser(
        prog = get_prog(),
        description  = 'Create a single-file version of an HTML file',
    )
    parser.add_argument('url')
    parser.add_argument(
        '-T', '--timeout',
        type=float,
        default=900.0,
        metavar= 'SECONDS',
        help='default connect and read timeout in seconds'
    )
    parser.add_argument(
        '--connect-timeout',
        type=float,
        metavar= 'SECONDS',
        help='default connect timeout in seconds (will override --timeout)'
    )
    parser.add_argument(
        '--read-timeout',
        type=float,
        metavar= 'SECONDS',
        help='default read timeout in seconds (will override --timeout)'
    )
    parser.add_argument(
        '--knockout',
        action= 'append',
        metavar= 'SELECTOR',
        help='knock out elements matching the given CSS selector'
    )

    args = parser.parse_args()

    timeout = args.timeout
    if (args.connect_timeout or args.read_timeout):
        timeout = (args.connect_timeout or timeout, args.read_timeout or timeout)

    print(freeze_to_string(args.url, timeout=timeout, knockouts=args.knockout))

    return 0

if __name__ == '__main__':
    sys.exit(main())



import pytest

import requests
from requests_testadapter import TestAdapter, TestSession

from frozen_soup import freeze_to_string

@pytest.fixture
def session() -> requests.Session:
    s = TestSession()
    s.mount('http://test/simple-string', TestAdapter(b'Alive!'))

    s.mount(
        "http://test/1x1.gif",
        TestAdapter(
            stream= open("tests/test_data/1x1.gif", "rb").read(),
            headers= {
                'Content-type': 'image/gif',
            },
        ),
    )
    s.mount("http://test/bad.gif", TestAdapter(b'BAD', status=404))
    s.mount("http://test/html-one-image", TestAdapter(b'<img src="/1x1.gif">'))
    s.mount("http://test/html-bad-image", TestAdapter(b'<img src="/bad.gif">'))

    s.mount(
        "http://test/style.css",
        TestAdapter(b'* { color: white }', headers= { 'Content-type': 'text/css' })
    )
    s.mount(
        "http://test/urls.css",
        TestAdapter(
            b'* { background: url(1x1.gif) } @media screen { background: url(1x1.gif) }',
            headers = { 'Content-type': 'text/css' }
        )
    )

    s.mount("http://test/html-link-icon", TestAdapter(b'<link rel="icon" href="1x1.gif">'))
    s.mount("http://test/html-link-style", TestAdapter(b'<link rel="stylesheet" href="style.css">'))
    s.mount(
        "http://test/html-link-style-urls",
        TestAdapter(b'<link rel="stylesheet" href="urls.css">')
    )

    s.mount(
        "http://test/code.js",
        TestAdapter(b'/* Code! */', headers= { 'Content-type': 'application/javascript' })
    )

    s.mount("http://test/html-script", TestAdapter(b'<script src="code.js"></script>'))

    # Not really data, but we're just testing to see it doesn't get mangled
    s.mount("http://test/already-data-url", TestAdapter(b'<img src="data:foo;bar:baz">'))

    return s


def test_freeze_to_string(session):
    out = freeze_to_string('http://test/simple-string', session)

    assert out == 'Alive!'


@pytest.fixture
def data_url():
    return "data:image/gif;base64,R0lGODlhAQABAIAAAMRUFwAAACH5BAAAAAAALAAAAAABAAEAAAICRAEAOw=="

def test_single_image(session, data_url):
    out = freeze_to_string('http://test/html-one-image', session)

    assert out == f'<img src="{data_url}">'

def test_single_image_as_xhtml(session, data_url):
    out = freeze_to_string('http://test/html-one-image', session, formatter='html')

    assert out == f'<img src="{data_url}"/>'

def test_bad_image(session):
    with pytest.raises(Exception):
        out = freeze_to_string('http://test/html-bad-image', session)

def test_link_icon(session, data_url):
    out = freeze_to_string('http://test/html-link-icon', session)

    assert out == f'<link href="{data_url}" rel="icon">'

def test_link_style(session):
    out = freeze_to_string('http://test/html-link-style', session)

    assert out == '<style>* { color: white }</style>'

def test_link_style_urls(session, data_url):
    out = freeze_to_string('http://test/html-link-style-urls', session)

    assert out == f'<style>* {{ background: url({data_url}) }} @media screen {{ background: url({data_url}) }}</style>'

def test_script(session):
    out = freeze_to_string('http://test/html-script', session)

    assert out == '<script>/* Code! */</script>';

def test_already_data_url(session):
    out = freeze_to_string('http://test/already-data-url', session)

    assert out == '<img src="data:foo;bar:baz">';



import pytest

import requests
from requests_testadapter import TestAdapter, TestSession

from frozen_soup import freeze_to_string

@pytest.fixture
def session() -> requests.Session:
    s = TestSession()

    s.mount("http://based/content", TestAdapter(
        b'/* WONTON */',
        headers= { 'Content-type' : 'text/plain' }
    ))

    s.mount("http://test/img", TestAdapter(b'<base href="http://based"><img src="/content">'))
    s.mount("http://test/script", TestAdapter(
        b'<base href="http://based"><script src="/content"></script>'
    ))
    s.mount("http://test/stylesheet", TestAdapter(
        b'<base href="http://based"><link rel="stylesheet" href="/content">'
    ))

    s.mount("http://test/extra-base", TestAdapter(
        b'<base target="_blank"><base href="http://based"><script src="/content"></script>'
    ))

    return s

def test_base_img(session):
    out = freeze_to_string('http://test/img', session)
    assert out == '<base href="http://based"><img src="data:text/plain;base64,LyogV09OVE9OICov">'

def test_base_script(session):
    out = freeze_to_string('http://test/script', session)
    assert out == '<base href="http://based"><script>/* WONTON */</script>'

def test_base_stylesheet(session):
    out = freeze_to_string('http://test/stylesheet', session)
    assert out == '<base href="http://based"><style>/* WONTON */</style>'

def test_base_extra_base(session):
    out = freeze_to_string('http://test/extra-base', session)
    assert out == '<base target="_blank"><base href="http://based"><script>/* WONTON */</script>'



import pytest

import requests
from requests_testadapter import TestAdapter, TestSession

from urllib3.exceptions import ConnectTimeoutError, ReadTimeoutError

from frozen_soup import freeze_to_string

class TimeoutTestAdapter(TestAdapter):
    def __init__(self, stream, timeout=None, status=200, headers=None):
        self.stream = stream
        self.status = status
        self.headers = headers or {}
        self.timeout = timeout
        super().__init__(stream, status, headers)

    def send(self, request, stream=False, timeout=None,
             verify=True, cert=None, proxies=None):

        if type(timeout) == int:
            if (timeout < self.timeout):
                raise ConnectTimeoutError
        elif type(timeout) == tuple:
            if (timeout[0] < self.timeout):
                raise ConnectTimeoutError
            elif (timeout[1] < self.timeout):
                raise ReadTimeoutError("(dummy)", request.url, "Read timeout!")

        return super().send(request, stream, timeout, verify, cert, proxies)

@pytest.fixture
def session() -> requests.Session:
    s = TestSession()

    s.mount("http://test/long-timeout",  TimeoutTestAdapter(b'DATA', timeout= 1200))
    s.mount("http://test/short-timeout", TimeoutTestAdapter(b'DATA', timeout= 3))

    s.mount("http://test/img-contains-long-timeout", TestAdapter(b'<img src="long-timeout">'))
    s.mount("http://test/img-contains-short-timeout", TestAdapter(b'<img src="short-timeout">'))

    s.mount(
        "http://test/link-contains-long-timeout",
        TestAdapter(b'<link rel="icon" href="long-timeout">')
    )
    s.mount(
        "http://test/link-contains-short-timeout",
        TestAdapter(b'<link rel="icon" href="short-timeout">')
    )

    s.mount(
        "http://test/script-contains-long-timeout",
        TestAdapter(b'<script src="long-timeout"></script>')
    )
    s.mount(
        "http://test/script-contains-short-timeout",
        TestAdapter(b'<script src="short-timeout"></script>')
    )

    s.mount(
        "http://test/long-timeout-css",
        TimeoutTestAdapter(b'* { background: white; }', timeout= 1200)
    )
    s.mount(
        "http://test/short-timeout-css",
        TimeoutTestAdapter(b'* { background: white; }', timeout= 3)
    )

    s.mount(
        "http://test/link-stylesheet-contains-long-timeout",
        TestAdapter(b'<link rel="stylesheet" href="long-timeout-css">')
    )
    s.mount(
        "http://test/link-stylesheet-contains-short-timeout",
        TestAdapter(b'<link rel="stylesheet" href="short-timeout-css">')
    )

    return s

def test_timeout_okay(session):
    out = freeze_to_string('http://test/short-timeout', session, timeout= 100)
    assert out == 'DATA'

def test_timeout_raised(session):
    with pytest.raises(ConnectTimeoutError):
        out = freeze_to_string('http://test/long-timeout', session, timeout= 100)

def test_timeout_okay_on_img(session):
    out = freeze_to_string('http://test/img-contains-short-timeout', session, timeout= 100)
    assert out == '<img src="data:None;base64,REFUQQ==">'

def test_timeout_raised_on_img(session):
    with pytest.raises(ConnectTimeoutError):
        out = freeze_to_string('http://test/img-contains-long-timeout', session, timeout= 100)

def test_timeout_tuple_okay(session):
    out = freeze_to_string('http://test/short-timeout', session, timeout= (100,100))
    assert out == 'DATA'

def test_timeout_tuple_connect_raised(session):
    with pytest.raises(ConnectTimeoutError):
        out = freeze_to_string('http://test/long-timeout', session, timeout= (100, 1200))

def test_timeout_tuple_read_raised(session):
    with pytest.raises(ReadTimeoutError):
        out = freeze_to_string('http://test/long-timeout', session, timeout= (1200, 100))

def test_timeout_okay_on_link(session):
    out = freeze_to_string('http://test/link-contains-short-timeout', session, timeout= 100)
    assert out == '<link href="data:None;base64,REFUQQ==" rel="icon">'

def test_timeout_raised_on_link(session):
    with pytest.raises(ConnectTimeoutError):
        out = freeze_to_string('http://test/link-contains-long-timeout', session, timeout= 100)

def test_timeout_okay_on_script(session):
    out = freeze_to_string('http://test/script-contains-short-timeout', session, timeout= 100)
    assert out == '<script>DATA</script>'

def test_timeout_raised_on_script(session):
    with pytest.raises(ConnectTimeoutError):
        out = freeze_to_string('http://test/script-contains-long-timeout', session, timeout= 100)

def test_timeout_okay_on_link_stylesheet(session):
    out = freeze_to_string('http://test/link-stylesheet-contains-short-timeout', session, timeout= 100)
    assert out == '<style>* { background: white; }</style>'

def test_timeout_raised_on_link_stylesheet(session):
    with pytest.raises(ConnectTimeoutError):
        out = freeze_to_string('http://test/link-stylesheet-contains-long-timeout', session, timeout= 100)



import pytest

import requests
from requests_testadapter import TestAdapter, TestSession

from frozen_soup import freeze_to_string

@pytest.fixture
def session() -> requests.Session:
    s = TestSession()

    s.mount("http://test/content", TestAdapter(
        b'/* WONTON */',
        headers= { 'Content-type' : 'text/plain' }
    ))

    s.mount(
        "http://test/html",
        TestAdapter(b'<i class="ko">pow!</i><img src="/content">')
    )
    s.mount(
        "http://test/multiple",
        TestAdapter(b'<i class="ko">pow!</i><b class="ko">bang!</b><img src="/content">')
    )
    s.mount(
        "http://test/bad-img",
        TestAdapter(b'<i class="ko">pow!</i><img src="/error">')
    )

    return s

def test_knockout(session):
    out = freeze_to_string('http://test/html', session, knockouts=['.ko'])
    assert out == '<img src="data:text/plain;base64,LyogV09OVE9OICov">'

def test_knockout_multiple_elements(session):
    out = freeze_to_string('http://test/multiple', session, knockouts=['.ko'])
    assert out == '<img src="data:text/plain;base64,LyogV09OVE9OICov">'

def test_knockout_multiple_selectors(session):
    out = freeze_to_string('http://test/multiple', session, knockouts=['i', 'b'])
    assert out == '<img src="data:text/plain;base64,LyogV09OVE9OICov">'

# if the knockout doesn't kill the <img> we'll get an exception
def test_knockout_img(session):
    out = freeze_to_string('http://test/bad-img', session, knockouts=['img'])
    assert out == '<i class="ko">pow!</i>'
